#include <stdio.h>
#include "boolean.h"
#include "archgtus.h"
#ifdef DEBUG
#include <assert.h>
#endif

void dec_tree(gt_mimx_data *current_mimx_data, gt_score *chosen_score
#ifdef PRUNING
	      , gt_score *oppbest
#endif
);

void Minimax(gt_data *data, gt_move *chosen_move,int parity) {
  /* in beginning range, out chosen_possibility, chosen_score */
  gt_mimx_data current_mimx_data;
  gt_score temp_score,best_score;
  gt_move temp_move;
  gt_unmove temp_unmove;
  boolean flag=false;
#ifndef FAST
  gt_moves move_array;
  int i;
#endif

#ifdef PRUNING
  boolean cutoff=false;
#endif

  current_mimx_data.depth=0;
  current_mimx_data.parity=parity;
#ifdef DEBUG
  current_mimx_data.node_count=1;
#endif
  current_mimx_data.current_data=*data;
  if ((current_mimx_data.depth+current_mimx_data.parity)%2==1)
    best_score = POS_INFINITY;
  else
    best_score = NEG_INFINITY;
#ifdef FAST
#ifdef PRUNING
  while(Gt_move(&current_mimx_data,&temp_move,&temp_unmove) && !cutoff)
#else
  while(Gt_move(&current_mimx_data,&temp_move,&temp_unmove))
#endif
#else
  Gt_generate_moves(&current_mimx_data,&move_array);
#ifdef PRUNING
  for (i=0;i<move_array.num_moves && !cutoff;i++)
#else
  for (i=0;i<move_array.num_moves;i++)
#endif
#endif
    {
#ifndef FAST
      temp_move=move_array.mlist[i];
      Gt_move(&current_mimx_data,&temp_move,&temp_unmove);
#endif
      current_mimx_data.depth++;
#ifdef PRUNING
      dec_tree(&current_mimx_data,&temp_score,&best_score);
#else
      dec_tree(&current_mimx_data,&temp_score);
#endif
      Gt_unmove(&current_mimx_data,&temp_unmove);
      current_mimx_data.depth--;
#ifdef DEBUG
      printf("at root score = %d  best = %d nodes = %d\n",(int)temp_score,(int)best_score,current_mimx_data.node_count);
      /*Gt_print(current_mimx_data);*/
#endif
      if ((current_mimx_data.depth+current_mimx_data.parity)%2==0){
	if (temp_score>best_score) {
	  *chosen_move = temp_move;
	  best_score = temp_score;
#ifdef PRUNING
	  if (best_score==POS_INFINITY)
	    cutoff = true;
#endif
	}
      }
      else
	if (temp_score<best_score) {
	  *chosen_move = temp_move;
	  best_score = temp_score;
#ifdef PRUNING
	  if (best_score==NEG_INFINITY)
	    cutoff = true;
#endif
	}
      if (!flag){
	*chosen_move=temp_move;
	flag=true;
      }
    }
#ifdef DEBUG
  if (!flag)
    printf("warning, no moves given to root node\n");
#endif
}

void dec_tree(gt_mimx_data *current_mimx_data,gt_score *chosen_score
#ifdef PRUNING
	      , gt_score *oppbest
#endif
) {
  gt_score temp_score,best_score;
#ifdef PRUNING
  boolean cutoff=false;
#endif
  gt_move temp_move;
  gt_unmove temp_unmove;
#ifndef FAST
  gt_moves move_array;
  int i;
#endif
#ifdef DEBUG
  int node=current_mimx_data->node_count++;
#endif
  if ((current_mimx_data->depth+current_mimx_data->parity)%2==1)
    best_score = POS_INFINITY;
  else
    best_score = NEG_INFINITY;
#ifdef FAST
#ifdef PRUNING
  while(Gt_move(current_mimx_data,&temp_move,&temp_unmove) && !cutoff)
#else
  while(Gt_move(current_mimx_data,&temp_move,&temp_unmove))
#endif
#else
  Gt_generate_moves(current_mimx_data,&move_array);
#ifdef PRUNING
  for (i=0;i<move_array.num_moves && !cutoff;i++)
#else
  for (i=0;i<move_array.num_moves;i++)
#endif
#endif
    {
#ifndef FAST
      temp_move=move_array.mlist[i];
      Gt_move(current_mimx_data,&temp_move,&temp_unmove);
#endif
      current_mimx_data->depth++;
#ifdef PRUNING
      dec_tree(current_mimx_data,&temp_score,&best_score);
#else
      dec_tree(current_mimx_data,&temp_score);
#endif
      Gt_unmove(current_mimx_data,&temp_unmove);
      current_mimx_data->depth--;
#ifdef DEBUG
      //printf("node = %d",node);/* depth = %d score = %d best = %d nodes = %d\n",node,current_mimx_data->depth,(int)temp_score,(int)best_score,current_mimx_data->node_count);
      /*Gt_print(current_mimx_data);*/
#endif
      if ((current_mimx_data->depth+current_mimx_data->parity)%2==0){
	if (temp_score>best_score) {
	  best_score = temp_score;
#ifdef PRUNING
	  if (best_score>=*oppbest)
	      cutoff = true;
#endif
	}
      }
      else
	if (temp_score<best_score) {
	  best_score = temp_score;
#ifdef PRUNING
	  if (best_score<=*oppbest)
	    cutoff = true;
#endif
	}
    }
  Gt_evaluate(current_mimx_data,chosen_score,best_score);
}
